class TestInvalidURI(object):
    def _makeOne(self, *args, **kwargs):
        from plaster.exceptions import InvalidURI
        return InvalidURI(*args, **kwargs)

    def test_it(self):
        exc = self._makeOne('foo')
        assert isinstance(exc, ValueError)
        assert exc.message == 'Unable to parse config_uri "foo".'
        assert exc.uri == 'foo'

    def test_it_overrides_message(self):
        exc = self._makeOne('foo', message='bar')
        assert isinstance(exc, ValueError)
        assert exc.message == 'bar'
        assert exc.uri == 'foo'


class TestLoaderNotFound(object):
    def _makeOne(self, *args, **kwargs):
        from plaster.exceptions import LoaderNotFound
        return LoaderNotFound(*args, **kwargs)

    def test_it(self):
        exc = self._makeOne('foo')
        assert isinstance(exc, ValueError)
        assert exc.scheme == 'foo'
        assert exc.protocols is None
        assert exc.message == (
            'Could not find a matching loader for the scheme "foo".')

    def test_it_with_protocol(self):
        exc = self._makeOne('foo', ['wsgi'])
        assert isinstance(exc, ValueError)
        assert exc.scheme == 'foo'
        assert exc.protocols == ['wsgi']
        assert exc.message == (
            'Could not find a matching loader for the scheme "foo", '
            'protocol "wsgi".')

    def test_it_with_multiple_protocols(self):
        exc = self._makeOne('foo', ['wsgi', 'qt'])
        assert isinstance(exc, ValueError)
        assert exc.scheme == 'foo'
        assert exc.protocols == ['wsgi', 'qt']
        assert exc.message == (
            'Could not find a matching loader for the scheme "foo", '
            'protocol "wsgi, qt".')

    def test_it_overrides_message(self):
        exc = self._makeOne('foo', message='bar')
        assert isinstance(exc, ValueError)
        assert exc.scheme == 'foo'
        assert exc.protocols is None
        assert exc.message == 'bar'


class TestMultipleLoadersFound(object):
    def _makeOne(self, *args, **kwargs):
        from plaster.exceptions import MultipleLoadersFound
        return MultipleLoadersFound(*args, **kwargs)

    def test_it(self):
        dummy1 = DummyLoaderInfo('dummy1')
        dummy2 = DummyLoaderInfo('dummy2')
        exc = self._makeOne('https', [dummy1, dummy2])
        assert isinstance(exc, ValueError)
        assert exc.message == (
            'Multiple plaster loaders were found for scheme "https". '
            'Please specify a more specific config_uri. Matched loaders: '
            'dummy1, dummy2')
        assert exc.scheme == 'https'
        assert exc.protocols is None
        assert exc.loaders == [dummy1, dummy2]

    def test_it_with_protocol(self):
        dummy1 = DummyLoaderInfo('dummy1')
        dummy2 = DummyLoaderInfo('dummy2')
        exc = self._makeOne('https', [dummy1, dummy2], protocols=['wsgi'])
        assert isinstance(exc, ValueError)
        assert exc.message == (
            'Multiple plaster loaders were found for scheme "https", '
            'protocol "wsgi". Please specify a more specific config_uri. '
            'Matched loaders: dummy1, dummy2')
        assert exc.scheme == 'https'
        assert exc.protocols == ['wsgi']
        assert exc.loaders == [dummy1, dummy2]

    def test_it_with_multiple_protocols(self):
        dummy1 = DummyLoaderInfo('dummy1')
        dummy2 = DummyLoaderInfo('dummy2')
        exc = self._makeOne('https', [dummy1, dummy2],
                            protocols=['wsgi', 'qt'])
        assert isinstance(exc, ValueError)
        assert exc.message == (
            'Multiple plaster loaders were found for scheme "https", '
            'protocol "wsgi, qt". Please specify a more specific '
            'config_uri. Matched loaders: dummy1, dummy2')
        assert exc.scheme == 'https'
        assert exc.protocols == ['wsgi', 'qt']
        assert exc.loaders == [dummy1, dummy2]

    def test_it_overrides_message(self):
        dummy = object()
        exc = self._makeOne('https', [dummy], message='foo')
        assert isinstance(exc, ValueError)
        assert exc.message == 'foo'
        assert exc.scheme == 'https'
        assert exc.protocols is None
        assert exc.loaders == [dummy]


class DummyLoaderInfo(object):
    def __init__(self, scheme):
        self.scheme = scheme



from setuptools import setup, find_packages

setup(
    name="app2",
    version="1.0",
    packages=find_packages(),
    entry_points={
        'plaster.loader_factory': [
            'dup=app2.loaders:DuplicateLoader',
            'yaml+bar=app2.loaders:YAMLBarLoader',
        ],
    },
)



def test_wsgi_protocol():
    from plaster.protocols import IWSGIProtocol

    class WSGILoader(IWSGIProtocol): # pragma: no cover
        def get_wsgi_app(self, name=None, defaults=None):
            def app(environ, start_response):
                start_response(b'200 OK', [(b'Content-Type', b'text/plain')])
                return [b'hello world']
            return app

        def get_wsgi_app_settings(self, name=None, defaults=None):
            settings = defaults.copy() if defaults else {}
            return settings

        def get_wsgi_filter(self, name=None, defaults=None):
            def filter(app):
                def wrapper(environ, start_response):
                    return app(environ, start_response)
                return wrapper
            return filter

        def get_wsgi_server(self, name=None, defaults=None):
            def server(app):
                from wsgiref.simple_server import make_server
                server = make_server('0.0.0.0', 8080, app)
                server.serve_forever()

    WSGILoader()



from setuptools import setup, find_packages

setup(
    name="app1",
    version="1.0",
    packages=find_packages(),
    entry_points={
        'plaster.loader_factory': [
            'conf=app1.loaders:ConfLoader',
            'file+conf=app1.loaders:ConfLoader',
            'ini=app1.loaders:INIWSGILoader',
            'file+ini=app1.loaders:INIWSGILoader',
            'file+yaml=app1.loaders:YAMLLoader',
            'yaml+foo=app1.loaders:YAMLFooLoader',
            'dup=app1.loaders:DuplicateLoader',
            'bad=app1.loaders:BadLoader',
            'broken=app1.loaders.BadLoader',
        ],
        'plaster.wsgi_loader_factory': [
            'ini=app1.loaders:INIWSGILoader',
            'file+ini=app1.loaders:INIWSGILoader',
        ],
        'plaster.dummy1_loader_factory': [
            'ini=app1.loaders:INIWSGILoader',
            'file+ini=app1.loaders:INIWSGILoader',
        ],
        'plaster.dummy2_loader_factory': [
            'ini=app1.loaders:INILoader',
            'file+ini=app1.loaders:INILoader',
        ],
        'other.loader': [
            'ini=app1.loaders:WontBeLoaded',
        ],
    },
)



import os.path
import pytest


class TestURL(object):
    def _callFUT(self, uri):
        from plaster.uri import parse_uri
        return parse_uri(uri)

    def test_relative_path(self):
        uri = self._callFUT('development.ini')
        assert uri.scheme == 'file+ini'
        assert uri.path == 'development.ini'
        assert uri.options == {}
        assert uri.fragment == ''

    def test_absolute_path(self):
        path = os.path.abspath('/path/to/development.ini')
        uri = self._callFUT(path)
        assert uri.scheme == 'file+ini'
        assert uri.path == path
        assert uri.options == {}
        assert uri.fragment == ''

    def test_absolute_path_with_fragment(self):
        path = os.path.abspath('/path/to/development.ini')
        uri = self._callFUT(path + '?a=b&c=d#main')
        assert uri.scheme == 'file+ini'
        assert uri.path == path
        assert uri.options == {'a': 'b', 'c': 'd'}
        assert uri.fragment == 'main'

    def test_url(self):
        uri = self._callFUT('redis://username@password:localhost/foo?a=b#main')
        assert uri.scheme == 'redis'
        assert uri.path == 'username@password:localhost/foo'
        assert uri.options == {'a': 'b'}
        assert uri.fragment == 'main'

    def test_url_for_file(self):
        uri = self._callFUT('pastedeploy+ini://development.ini')
        assert uri.scheme == 'pastedeploy+ini'
        assert uri.path == 'development.ini'
        assert uri.options == {}
        assert uri.fragment == ''

    def test_missing_scheme(self):
        from plaster.exceptions import InvalidURI
        with pytest.raises(InvalidURI):
            self._callFUT('foo')

    def test___str__(self):
        uri = self._callFUT('development.ini')
        assert str(uri) == 'file+ini://development.ini'

    def test___str___with_options(self):
        uri = self._callFUT('development.ini?a=b&c=d')
        assert str(uri) == 'file+ini://development.ini?a=b&c=d'

    def test___str___with_fragment(self):
        uri = self._callFUT('development.ini#main')
        assert str(uri) == 'file+ini://development.ini#main'

    def test___repr___(self):
        uri = self._callFUT('development.ini#main')
        assert repr(uri) == 'PlasterURL(\'file+ini://development.ini#main\')'

    def test_returns_same_instance(self):
        uri1 = self._callFUT('development.ini')
        uri2 = self._callFUT(uri1)
        assert uri1 is uri2

    def test_colon_prefix_scheme(self):
        uri = self._callFUT('egg:myapp#main')
        assert uri.scheme == 'egg'
        assert uri.path == 'myapp'
        assert uri.fragment == 'main'

    def test_only_scheme(self):
        uri = self._callFUT('egg:')
        assert uri.scheme == 'egg'
        assert uri.path == ''
        assert uri.options == {}
        assert uri.fragment == ''


def test_default_url_values():
    from plaster.uri import PlasterURL
    url = PlasterURL('foo')
    assert url.scheme == 'foo'
    assert url.path == ''
    assert url.options == {}
    assert url.fragment == ''



import pytest

@pytest.mark.usefixtures('fake_packages')
class Test_get_loader(object):
    def _callFUT(self, *args, **kwargs):
        from plaster.loaders import get_loader
        return get_loader(*args, **kwargs)

    def test_simple_uri(self):
        loader = self._callFUT('development.conf')
        assert loader.entry_point_key == 'conf'

    def test_scheme_uri(self):
        loader = self._callFUT('conf://development.conf')
        assert loader.entry_point_key == 'conf'

    def test_scheme_uri_for_pkg(self):
        loader = self._callFUT('app1+conf://')
        assert loader.entry_point_key == 'conf'

    def test_path_with_extension(self):
        loader = self._callFUT('development.ini')
        assert loader.entry_point_key == 'ini+wsgi'

    def test_path_with_extension_and_protocol(self):
        loader = self._callFUT('development.ini', protocols=['wsgi'])
        assert loader.entry_point_key == 'ini+wsgi'

    def test_dup(self):
        from plaster.exceptions import MultipleLoadersFound
        with pytest.raises(MultipleLoadersFound):
            self._callFUT('dup://development.ini')

    def test_dedup_app1(self):
        loader = self._callFUT('app1+dup://development.ini')
        assert loader.entry_point_key == 'app1+dup'

    def test_dedup_app2(self):
        loader = self._callFUT('app2+dup://development.ini')
        assert loader.entry_point_key == 'app2+dup'

    def test_other_groups(self):
        from plaster.exceptions import LoaderNotFound
        with pytest.raises(LoaderNotFound):
            self._callFUT('other-scheme://development.ini')

    def test_bad(self):
        from app1.loaders import BadLoader
        loader = self._callFUT('bad:development')
        assert isinstance(loader, BadLoader)

    def test_it_broken(self):
        with pytest.raises(Exception):
            self._callFUT('development.broken')

    def test_it_notfound(self):
        from plaster.exceptions import LoaderNotFound
        with pytest.raises(LoaderNotFound):
            self._callFUT('development.notfound')

    def test_fallback_non_pkg_scheme(self):
        loader = self._callFUT('yaml+bar://development.yml')
        assert loader.entry_point_key == 'yaml+bar'


@pytest.mark.usefixtures('fake_packages')
class Test_find_loaders(object):
    def _callFUT(self, *args, **kwargs):
        from plaster.loaders import find_loaders
        return find_loaders(*args, **kwargs)

    def test_simple_uri(self):
        loaders = self._callFUT('conf')
        assert len(loaders) == 1
        assert loaders[0].scheme == 'app1+conf'
        loader = loaders[0].load('development.conf')
        assert loader.entry_point_key == 'conf'

    def test_case_insensitive_scheme(self):
        loaders = self._callFUT('CONF')
        assert len(loaders) == 1
        assert loaders[0].scheme == 'app1+conf'
        loader = loaders[0].load('development.conf')
        assert loader.entry_point_key == 'conf'

    def test_scheme_specific_uri(self):
        loaders = self._callFUT('ini')
        assert len(loaders) == 1
        assert loaders[0].scheme == 'app1+ini'
        loader = loaders[0].load('development.ini')
        assert loader.entry_point_key == 'ini+wsgi'

    def test_multiple_yaml_loaders(self):
        loaders = self._callFUT('dup')
        assert len(loaders) == 2
        schemes = set([l.scheme for l in loaders])
        assert 'app1+dup' in schemes
        assert 'app2+dup' in schemes

    def test_one_protocol(self):
        loaders = self._callFUT('ini', protocols=['wsgi'])
        assert len(loaders) == 1
        loader = loaders[0].load('development.ini')
        assert loader.entry_point_key == 'ini+wsgi'

    def test_multiple_protocols(self):
        loaders = self._callFUT('ini', protocols=['wsgi', 'dummy1'])
        assert len(loaders) == 1
        loader = loaders[0].load('development.ini')
        assert loader.entry_point_key == 'ini+wsgi'

    def test_multiple_incompatible_protocols(self):
        loaders = self._callFUT('ini', protocols=['wsgi', 'dummy2'])
        assert len(loaders) == 0

    def test_it_notfound(self):
        loaders = self._callFUT('notfound')
        assert len(loaders) == 0


@pytest.mark.usefixtures('fake_packages')
class Test_get_sections(object):
    def _callFUT(self, config_uri):
        from plaster.loaders import get_sections
        return get_sections(config_uri)

    def test_it(self):
        result = self._callFUT('development.ini')
        assert set(result) == set(['a', 'b'])

    def test_it_bad(self):
        with pytest.raises(Exception):
            self._callFUT('development.bad')


@pytest.mark.usefixtures('fake_packages')
class Test_get_settings(object):
    def _callFUT(self, config_uri, section=None, defaults=None):
        from plaster.loaders import get_settings
        return get_settings(config_uri, section=section, defaults=defaults)

    def test_it_explicit_a(self):
        result = self._callFUT('development.ini', 'a')
        assert result == {'foo': 'bar'}

    def test_it_explicit_b(self):
        result = self._callFUT('development.ini', 'b')
        assert result == {'baz': 'xyz'}

    def test_it_fragment(self):
        result = self._callFUT('development.ini#a')
        assert result == {'foo': 'bar'}

    def test_defaults(self):
        result = self._callFUT('development.ini', 'a', {'baz': 'foo'})
        assert result == {'foo': 'bar', 'baz': 'foo'}

    def test_invalid_section(self):
        result = self._callFUT('development.ini', 'c')
        assert result == {}

    def test_it_bad(self):
        with pytest.raises(Exception):
            self._callFUT('development.bad')


@pytest.mark.usefixtures('fake_packages')
class Test_setup_logging(object):
    def _makeOne(self, config_uri):
        from plaster.loaders import get_loader
        return get_loader(config_uri)

    def _callFUT(self, config_uri, defaults=None):
        from plaster.loaders import setup_logging
        return setup_logging(config_uri, defaults=defaults)

    def test_it(self):
        loader = self._makeOne('development.ini#a')
        loader.setup_logging()
        assert loader.logging_setup
        assert loader.logging_defaults is None

    def test_it_top_level(self):
        self._callFUT('development.ini#a')

    def test_it_bad(self):
        with pytest.raises(Exception):
            self._callFUT('bad://development.ini')



import plaster
from plaster.protocols import IWSGIProtocol

_SECTIONS = {
    'a': {
        'foo': 'bar',
    },
    'b': {
        'baz': 'xyz',
    },
}


class LoaderBase(plaster.ILoader):
    entry_point_key = None

    def __init__(self, uri):
        self.uri = uri

    def get_sections(self):
        return list(_SECTIONS.keys())

    def get_settings(self, section=None, defaults=None):
        if section is None:
            section = self.uri.fragment
        if defaults is not None:
            result = defaults.copy()
        else:
            result = {}
        try:
            result.update(_SECTIONS[section])
        except KeyError:
            pass
        return result

    def setup_logging(self, defaults=None):
        self.logging_setup = True
        self.logging_defaults = defaults


class ConfLoader(LoaderBase):
    entry_point_key = 'conf'


class INILoader(LoaderBase):
    entry_point_key = 'ini'


class INIWSGILoader(IWSGIProtocol, LoaderBase):
    entry_point_key = 'ini+wsgi'

    def get_wsgi_app(self, name=None, defaults=None):
        return 'wsgi app'

    def get_wsgi_app_settings(self, name=None, defaults=None):
        return {'a': 'b'}

    def get_wsgi_filter(self, name=None, defaults=None):
        return 'wsgi filter'

    def get_wsgi_server(self, name=None, defaults=None):
        return 'wsgi server'


class YAMLLoader(LoaderBase):
    entry_point_key = 'yaml'


class YAMLFooLoader(LoaderBase):
    entry_point_key = 'yaml+foo'


class DuplicateLoader(LoaderBase):
    entry_point_key = 'app1+dup'


class BadLoader(object):
    def __init__(self, uri):
        self.uri = uri


class WontBeLoaded(LoaderBase):
    entry_point_key = 'ini'
